import sys
import logging
from .state import CREATION_MODE

logger = logging.getLogger(__name__)

class Infrastructure(object):

    _wrapped = None
    _dependencies = None
    _category = None
    _original_creation_mode = None
    _locals = None
    _result = None

    def __init__(self, callable_, *args, **kwargs):
        self.__name__ = callable_.__name__ if hasattr(callable_, '__name__') else 'undefined'
        self.__doc__ = callable_.__doc__ if hasattr(callable_, '__doc__') else None
        self.__module__ = callable_.__module__ if hasattr(callable_, '__module__') else None

        self._wrapped = callable_
        self.environment = kwargs.get('environment', None)
        self.dependencies = kwargs.get('requires', None)

    def __repr__(self):
        return 'Infrastructure:' + self.__name__

    def __call__(self, *args, **kwargs):

        # Set the creation mode, if the object specifies one.
        self._set_creation_mode()

        # Define a source code profiler.
        def profiler(frame, event, arg):
            if event == 'return':
                self._locals = frame.f_locals.copy()

        # Activate the profiler on the next call, return or exception.
        sys.setprofile(profiler)
        try:
            # Trace the function call.
            self._result = self._wrapped(*args, **kwargs)
        finally:
            # Disable the source code profiler.
            sys.setprofile(None)

        # Reset the creation mode, if the object specifies one.
        self._reset_creation_mode()

        return self._result

    def __getattr__(self, attr):
            return self._locals[attr]

    def _set_creation_mode(self):
        global CREATION_MODE
        if self.category:
            self._original_creation_mode = CREATION_MODE
            CREATION_MODE = self.category
            logger.debug('Set CREATION_MODE to \'%s\'.' % self.category.title())

    def _reset_creation_mode(self):
        global CREATION_MODE
        if self.category:
            CREATION_MODE = self._original_creation_mode
            logger.debug('Set CREATION_MODE to \'%s\'.' % (self._original_creation_mode.title() \
                                                           if isinstance(self._original_creation_mode, str) \
                                                           else self._original_creation_mode))

    @property
    def dependencies(self):
        return self._dependencies

    @dependencies.setter
    def dependencies(self, dependencies):
        if dependencies:
            self._dependencies = set(dependencies)
            logger.debug('Set (%s) dependencies to (%s).' % (self, ', '.join(list(dependencies))))

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, category):
        self._category = category
        logger.debug('Set (%s) to \'%s\' Creation Mode.' % (self, category.title()))

    @property
    def resources(self):
        return self._locals

    @property
    def result(self):
        return self._result



import os
import tarfile
import random
import logging
import boto
from .state import config

logger = logging.getLogger(__name__)

def connect_s3():
    logger.debug('Connecting to the Amazon Simple Storage Service (Amazon S3).')
    s3 = boto.connect_s3(aws_access_key_id=config['AWS_ACCESS_KEY_ID'],
                         aws_secret_access_key=config['AWS_SECRET_ACCESS_KEY'])
    logger.debug('Connected to Amazon S3.')

    return s3

def create_bucket():
    s3_connection = connect_s3()
    s3_bucket_name = '-'.join(['s3', config['PROJECT_NAME'], config['ENVIRONMENT'], '%x' % random.randrange(2**32)])
    lifecycle_config = boto.s3.lifecycle.Lifecycle()
    lifecycle_config.add_rule(status='Enabled', expiration=1)
    bucket = s3_connection.lookup(s3_bucket_name)
    if not bucket:
        try:
            logger.info('Creating S3 bucket (%s).' % s3_bucket_name)
            bucket = s3_connection.create_bucket(s3_bucket_name, location=boto.s3.connection.Location.DEFAULT, policy='private')
            bucket.configure_lifecycle(lifecycle_config)
            logger.info('Created S3 bucket (%s).' % s3_bucket_name)
        except boto.exception.S3CreateError as error:
            logger.error('Could not create S3 bucket (%s) due to Error %s: %s.' % (s3_bucket_name, error.status, error.reason))
    else:
        bucket = create_bucket()

    return bucket

def add_object(bucket, obj):
    key = bucket.new_key(obj)
    key.set_contents_from_filename(obj, policy='private')

def get_bucket_policy(bucket):
    arns = ['"arn:aws:s3:::'+bucket.name+'/'+key.name+'"' for key in bucket.get_all_keys()]
    policy = """{
        "Version": "2012-10-17",
        "Statement": [{
            "Effect": "Allow",
            "Action": [
                "s3:GetObject"
            ],
            "Resource": [%s]
        }]
    }""" % ','.join(arns)
    return policy



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""main.py: Provision AWS environments quickly."""

import os
import sys
import types
from string import Template
import logging
import importlib
from .utils import parse_arguments
from .infrastructure import Infrastructure
from .state import ready

__author__ = 'Jared Contrascere'
__copyright__ = 'Copyright 2015, LibreTees, LLC. All rights reserved.'
__license__ = 'GPLv3'

logger = logging.getLogger(__name__)

def get_script(region, s3bucket, s3object, s3object2, filename='user-data.sh'):
    template = open(filename).read()
    return Template(template).substitute(
        region=region,
        s3bucket=s3bucket,
        s3object=s3object,
        s3object2=s3object2
    )

def load_skyfile(path='./skyfile.py', module_name='skyfile'):
    # Import Skyfile (Only works in Python 3.3+).
    loader = importlib.machinery.SourceFileLoader(module_name, path)
    module = loader.load_module()
    logger.info('Loaded (%s) module from (%s).' % (module_name, path))

    return module

def load_infrastructure(module):

    infrastructure_objects = []

    imported_symbols = vars(module)

    if '__all__' in imported_symbols:
        # Obey the use of <module>.__all__, if it is present.
        imported_symbols = [(name, imported_symbols[name]) for name in \
                            imported_symbols if name in imported_symbols['__all__']]
    else:
        imported_symbols = imported_symbols.items()

    for symbol in imported_symbols:
        name, obj = symbol
        if isinstance(obj, Infrastructure):
            logger.debug('(%s) is a %s object imported from the (%s) module.' % (name, type(obj), module.__name__))
            infrastructure_objects.append(obj)
        elif isinstance(obj, types.ModuleType):
            logger.debug('(%s) module imported from the (%s) module.' % (name, module.__name__))
            infrastructure_objects += load_infrastructure(obj)

    return infrastructure_objects

def build_dependency_graph(nodes):

    # Create an empty graph.
    graph = []

    # Create a copy of the nodes list to work with.
    nodes = list(nodes)

    # Store the length of the nodes list, so that circular dependencies can be detected.
    original_count = len(nodes)

    # Determine independent nodes.
    independent_nodes = [node for node in nodes if not node.dependencies]

    if independent_nodes:
        # Remove independent nodes from the search space.
        for independent_node in independent_nodes:
            nodes.remove(independent_node)
            logger.debug('(%s) is ready.' % independent_node.__name__)

        # Prune independent node from all node dependencies.
        for name, dependencies in [(node.__name__, node.dependencies) for node in nodes if node.dependencies]:
            logger.debug('(%s) required: (%s).' % (name, ', '.join(list(dependencies))))
            resolved_dependencies = {independent_node.__name__ for independent_node in independent_nodes}
            dependencies.difference_update(resolved_dependencies)
            logger.debug('(%s) now requires: (%s).' % (name, ', '.join(list(dependencies))))

        # Repeat search for newly-independent nodes, if any are left.
        if nodes:
            graph = build_dependency_graph(nodes)

        # Build graph from the independent nodes to the most-dependent nodes.
        graph.insert(0, independent_nodes)

    # Check for circular dependencies.
    count = 0
    for depth in graph:
        for breadth in depth:
            count += 1
    if original_count != count:
        raise RuntimeError('Circular dependencies detected.')

    return graph

def main():
    parse_arguments()
    module = load_skyfile()
    infrastructure = load_infrastructure(module)

    graph = build_dependency_graph(infrastructure)

    for dependencies in graph:
        for dependency in dependencies:
            dependency()
            ready[dependency.__name__] = dependency

    # archive_name = '.'.join([s3.PROJECT_NAME, 'tar', 'gz'])
    # logger.info('Creating deployment archive (%s).' % archive_name)
    # s3.make_tarfile(archive_name, s3.PROJECT_DIRECTORY)
    # logger.info('Created deployment archive (%s).' % archive_name)

    # bootstrap_archive_name = '.'.join(['configure', s3.PROJECT_NAME, 'tar', 'gz'])
    # s3.make_tarfile(bootstrap_archive_name, 'deploy')

    # bucket = s3.create_bucket()
    # s3.add_object(bucket, archive_name)
    # s3.add_object(bucket, bootstrap_archive_name)
    # policy = s3.get_bucket_policy(bucket)

    # script = get_script('us-east-1', bucket.name, archive_name, bootstrap_archive_name)
    # script = ec2.install_package(script, 'python3-pip')
    # script = ec2.run(script, 'pip3 install virtualenv')
    # script = ec2.run(script, 'pip3 install virtualenvwrapper')

    # instance_profile_name = iam.create_role(policy)

if __name__ == '__main__':
    main()



from .state import ready
from .decorators import permanent, ephemeral, infrastructure
from .networking import create_network, create_subnets
from .compute import (get_instances, create_instances, terminate_instances,
                      create_nat_instances, create_security_group, create_load_balancer, register_instances)
from .database import create_database
from .security import upload_ssl_certificate


import functools
import logging
from .infrastructure import Infrastructure
from .state import CREATION_MODE, EPHEMERAL, PERMANENT

logger = logging.getLogger(__name__)

def ephemeral(*args, **kwargs):
    # Determine whether or not the decorator was invoked.
    invoked = bool(not args or args and not callable(args[0]) or kwargs)
    logger.debug('@ephemeral Decorator was invoked.' if invoked else '@ephemeral Decorator was not invoked.')

    if not invoked:
        if isinstance(args[0], Infrastructure):
            # Set a wrapped Infractructure object to EPHEMERAL creation mode.
            infrastructure = args[0]
            infrastructure.category = EPHEMERAL
            decorator = infrastructure
        else:
            # Wrap a function in a decorator that sets the infrastructure creation mode to EPHEMERAL.
            function, args = args[0], ()
            @functools.wraps(function)
            def decorator(*args, **kwargs):
                CREATION_MODE = EPHEMERAL
                return function(*args, **kwargs)
            logging.info('Decorated (%s) as an \'Ephemeral\' Creation Mode function.' % function.__name__)
    else:
        # Define a decorator that will wrap a function in an Ephemeral Infrastructure object.
        def decorator(function):
            infrastructure = Infrastructure(function, *args, **kwargs)
            infrastructure.category = EPHEMERAL
            logging.info('Decorated (%s) as an Ephemeral Infrastructure object.' % function.__name__)
            return infrastructure

    return decorator

def permanent(*args, **kwargs):
    # Determine whether or not the decorator was invoked.
    invoked = bool(not args or args and not callable(args[0]) or kwargs)
    logger.debug('@permanent Decorator was invoked.' if invoked else '@permanent Decorator was not invoked.')

    if not invoked:
        if isinstance(args[0], Infrastructure):
            # Set a wrapped Infractructure object to PERMANENT creation mode.
            infrastructure = args[0]
            infrastructure.category = PERMANENT
            decorator = infrastructure
        else:
            # Wrap the function in a decorator that sets the infrastructure creation mode to PERMANENT.
            function, args = args[0], ()
            @functools.wraps(function)
            def decorator(*args, **kwargs):
                CREATION_MODE = PERMANENT
                return function(*args, **kwargs)
            logging.info('Decorated (%s) as a \'Permanent\' Creation Mode function.' % function.__name__)
    else:
        # Define a decorator that will wrap a function in an Permanent Infrastructure object.
        def decorator(function):
            infrastructure = Infrastructure(function, *args, **kwargs)
            infrastructure.category = PERMANENT
            logging.info('Decorated (%s) as an Permanent Infrastructure object.' % function.__name__)
            return infrastructure

    return decorator

def infrastructure(*args, **kwargs):
    # Determine whether or not the decorator was invoked.
    invoked = bool(not args or args and not callable(args[0]) or kwargs)
    logger.debug('@infrastructure Decorator was invoked.' if invoked else '@infrastructure Decorator was not invoked.')

    if not invoked:
        function, args = args[0], ()

    # Define decorator function.
    def decorator(function):
        return Infrastructure(function, *args, **kwargs)

    # If invoked, return the decorator function, which be called immediately. Otherwise, return the decorated function.
    return decorator if invoked else decorator(function)



import os
import sys
import logging
from re import search, IGNORECASE
from argparse import ArgumentParser
from configparser import ConfigParser
from timeit import Timer
from boto import regioninfo
from .state import config

logger = logging.getLogger(__name__)

def get_closest_region(service='ec2', repetitions=1):
    regions = [region.name for region in regioninfo.get_regions(service) if 'gov' not in region.name and 'cn' not in region.name]

    latency = {}
    for region in regions:
        connection = Timer("h.request('GET', '/')",
                           "from http.client import HTTPSConnection; h=HTTPSConnection('ec2.%s.amazonaws.com')" % region)
        times = connection.repeat(repetitions, 1)
        avg_latency = sum(times)/float(len(times))
        latency[region] = avg_latency
        logger.info('Average latency to Amazon %s %s is %s' % (service.upper(), region, latency[region]))

    region = min(latency, key=latency.get)
    return region

def make_tarfile(output_filename, source_dir):
    logger.info('Archiving directory (%s).' % source_dir)
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname='.')
    logger.info('Created gzipped tarball (%s).' % output_filename)

def configure_logger(args):
    # Restrict the boto logger to the WARNING log level.
    if args.loglevel.upper() in ['DEBUG', 'INFO', 'WARNING']:
        logging.getLogger('boto').setLevel(logging.WARNING)

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)
    logging.basicConfig(level=numeric_level)

def parse_arguments():
    global config
    
    valid_arguments = True
    parser = ArgumentParser(description='Provision Django application environments.')
    parser.add_argument('command', metavar='<command>', action='store', help='Valid commands are [deploy]')
    parser.add_argument('-p', '--project', dest='directory', action='store', default=os.getcwd(),
                        help='set Django project directory')
    parser.add_argument('-env', '--environment', dest='environment', action='store', default='STAGING',
                        help='set desired deployment environment [STAGING, PRODUCTION]\n(default: staging)')
    parser.add_argument('-aws', '--account-id', dest='account_id', action='store', default=os.environ.get('AWS_ACCOUNT_ID'),
                        help='set AWS Account ID')
    parser.add_argument('-id', '--key-id', dest='key_id', action='store', default=os.environ.get('AWS_ACCESS_KEY_ID'),
                        help='set AWS Access Key ID')
    parser.add_argument('-k', '--secret', dest='key', action='store', default=os.environ.get('AWS_SECRET_ACCESS_KEY'),
                        help='set AWS Account Secret Access Key')
    parser.add_argument('-d', '--log', dest='loglevel', action='store', default='ERROR',
                        help='set log level [DEBUG, INFO, WARNING, ERROR, CRITICAL] (default: ERROR)')
    parser.add_argument('--dry', dest='dry_run', action='store_true', default=False,
                        help='perform a dry run')

    # Display help, if no command was supplied.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    try:
        assert args.loglevel.upper() in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
        if args.loglevel.upper() == 'DEBUG':
            print('DEBUG:', __name__, ':Log argument validated (%s).' % args.loglevel, sep='')
    except AssertionError:
        print('ERROR:', __name__, ':Invalid log option (%s).' % args.loglevel, file=sys.stderr, sep='')
        valid_arguments = False

    configure_logger(args)

    try:
        assert args.command.upper() in ['DEPLOY']
        logger.debug('Command argument validated (%s).' % args.command)
    except AssertionError:
        logger.error('Invalid command (%s).' % args.command)
        valid_arguments = False

    try:
        assert args.environment.upper() in ['STAGING', 'PROD', 'PRODUCTION']
        if args.environment.upper() == 'PRODUCTION': args.environment = 'PROD'
        logger.debug('Deployment environment argument validated (%s).' % args.environment)
    except AssertionError:
        logger.error('Invalid deployment environment (%s).' % args.environment)
        valid_arguments = False

    try:
        assert os.path.isdir(os.path.expanduser(args.directory))
        logger.debug('Django project directory argument validated (%s).' % args.directory)
    except AssertionError:
        logger.error('Invalid Django project directory (%s).' % args.directory)
        valid_arguments = False

    try:
        assert search(r'^\d{12}$', args.account_id)
        logger.debug('AWS Account ID argument validated (%s).' % args.account_id)
    except AssertionError:
        if len(args.account_id):
            logger.error('AWS Account ID must be exactly 12 digits (%s).' % args.account_id)
        else:
            logger.error('AWS Account ID not specified.')
        valid_arguments = False

    config_path = None
    if os.environ.get('BOTO_CONFIG'):
        config_path = os.path.expanduser(os.environ.get('BOTO_CONFIG'))
    elif os.path.exists(os.path.expanduser('~/.boto')):
        config_path = os.path.expanduser('~/.boto')
    elif os.path.exists(os.path.expanduser('~/.aws/credentials')):
        config_path = os.path.expanduser('~/.aws/credentials')
    elif os.path.exists('/etc/boto.cfg'):
        config_path = '/etc/boto.cfg'

    if config_path:
        config = ConfigParser()
        config.sections()
        try:
            logger.info('Reading configuration file (%s).' % config_path)
            config.read(config_path)
            key_id = config['Credentials']['aws_access_key_id']
            key = config['Credentials']['aws_secret_access_key']
        except:
            logger.error('Could not read configuration file (%s).' % config_path)

    try:
        args.key_id = args.key_id or key_id
        assert search(r'^[A-Z0-9]{20}$', args.key_id, IGNORECASE)
        logger.debug('AWS Access Key ID argument validated (%s).' % args.key_id)
    except AssertionError:
        if len(args.key_id):
            logger.error('AWS Access Key ID must contain 20 alphanumeric characters (%s).' % args.key_id)
        else:
            logger.error('AWS Access Key ID not specified.')
        valid_arguments = False

    try:
        args.key = args.key or key
        assert search(r'^[A-Z0-9/\+]{40}$', args.key, IGNORECASE)
        logger.debug('AWS Account Secret Access Key argument validated.')
    except AssertionError:
        if len(args.key):
            logger.error('AWS Account Secret Access Key must contain 40 alphanumeric characters and/or the following: /+ (%s).' \
                         % args.key)
        else:
            logger.error('AWS Account Secret Access Key not specified.')
        valid_arguments = False

    if not valid_arguments:
        logger.error('Invalid arguments given.')
        logger.error('Exiting...')
        sys.exit(1)

    config['PROJECT_NAME'] = os.path.abspath(os.path.expanduser(args.directory)).split(os.sep)[-1].lower()
    config['PROJECT_DIRECTORY'] = os.path.abspath(os.path.expanduser(args.directory)).lower()
    config['ENVIRONMENT'] = args.environment.lower()
    config['AWS_ACCOUNT_ID'] = args.account_id
    config['AWS_ACCESS_KEY_ID'] = args.key_id
    config['AWS_SECRET_ACCESS_KEY'] = args.key

    return args



import time
import random
import logging
import boto
from .state import config

logger = logging.getLogger(__name__)

def connect_iam():
    logger.debug('Connecting to the Amazon Identity and Access Management (Amazon IAM) service.')
    iam = boto.connect_iam(aws_access_key_id=config['AWS_ACCESS_KEY_ID'],
                           aws_secret_access_key=config['AWS_SECRET_ACCESS_KEY'])
    logger.debug('Connected to Amazon IAM.')

    return iam

def delete_role(role_name):
    # Connect to the Amazon Identity and Access Management (Amazon IAM) service.
    iam_connection = connect_iam()

    # Get a list of all Instance Profiles associated to the Role.
    instance_profiles = None
    try:
        response = iam_connection.list_instance_profiles_for_role(role_name)
        instance_profiles = response['list_instance_profiles_for_role_response']\
                                    ['list_instance_profiles_for_role_result']\
                                    ['instance_profiles']
    except boto.exception.BotoServerError as error:
        if error.status == 404: # Not Found
            logger.error('Error %s: %s. Role (%s) was not found. ' % (error.status, error.reason, role_name))

    # Get a list of all Role Policies associated to the Role.
    role_policies = None
    try:
        response = iam_connection.list_role_policies(role_name)
        role_policies = response['list_role_policies_response']\
                                ['list_role_policies_result']\
                                ['policy_names']
    except boto.exception.BotoServerError as error:
        if error.status == 404: # Not Found
            logger.error('Error %s: %s. Role (%s) was not found. ' % (error.status, error.reason, role_name))

    # Remove Role from Instance Profiles.
    if instance_profiles:
        instance_profile_names = [instance_profile['instance_profile_name'] for instance_profile in instance_profiles]
        for instance_profile_name in instance_profile_names:
            logger.info('Removing Role (%s) from Instance Profile (%s).' % (role_name, instance_profile_name))
            try:
                iam_connection.remove_role_from_instance_profile(instance_profile_name, role_name)
                logger.info('Removed Role (%s) from Instance Profile (%s).' % (role_name, instance_profile_name))
            except boto.exception.BotoServerError as error:
                if error.status == 400: # Bad Request
                    logger.error('Error %s: %s. Couldn\'t remove Role (%s) Instance Profile (%s).' % (error.status, error.reason, role_name, instance_profile_name))
                if error.status == 404: # Not Found
                    logger.error('Error %s: %s. Instance Profile (%s) was not found. ' % (error.status, error.reason, instance_profile_name))

            # Delete Instance Profile.
            logger.info('Deleting Instance Profile (%s).' % instance_profile_name)
            try:
                iam_connection.delete_instance_profile(instance_profile_name)
                logger.info('Deleted Instance Profile (%s).' % instance_profile_name)
            except boto.exception.BotoServerError as error:
                if error.status == 400: # Bad Request
                    logger.error('Error %s: %s. Couldn\'t delete Instance Profile (%s).' % (error.status, error.reason, instance_profile_name))
                if error.status == 404: # Not Found
                    logger.error('Error %s: %s. Instance Profile (%s) was not found. ' % (error.status, error.reason, instance_profile_name))

    # Delete Role Policies from Role.
    if role_policies:
        for role_policy in role_policies:
            logger.info('Deleting Role Policy (%s) from Role (%s).' % (role_policy, role_name))
            try:
                iam_connection.delete_role_policy(role_name, role_policy)
                logger.info('Deleted Role Policy (%s) from Role (%s).' % (role_policy, role_name))
            except boto.exception.BotoServerError as error:
                if error.status == 400: # Bad Request
                    logger.error('Error %s: %s. Couldn\'t delete Policy (%s) Role (%s).' % (error.status, error.reason, role_policy, role_name))
                if error.status == 404: # Not Found
                    logger.error('Error %s: %s. Role (%s) was not found. ' % (error.status, error.reason, role_name))

    # Delete Role.
    logger.info('Deleting Role (%s).' % role_name)
    try:
        iam_connection.delete_role(role_name)
        logger.info('Deleted Role (%s).' % role_name)
    except boto.exception.BotoServerError as error:
        if error.status == 400: # Bad Request
            logger.error('Error %s: %s. Couldn\'t delete Role (%s).' % (error.status, error.reason, role_name))
        if error.status == 404: # Not Found
            logger.error('Error %s: %s. Role (%s) was not found. ' % (error.status, error.reason, role_name))

def create_role(inline_policies):
    # Connect to the Amazon Identity and Access Management (Amazon IAM) service.
    iam_connection = connect_iam()

    # Convert Inline Policies parameter to list, if necessary.
    if isinstance(inline_policies, str):
        inline_policies = [inline_policies]

    # Create Role.
    role_name = '-'.join(['role', config['PROJECT_NAME'], config['ENVIRONMENT']])
    delete_role(role_name)
    logger.info('Creating Role (%s).' % role_name)
    role = iam_connection.create_role(role_name)
    logger.info('Created Role (%s).' % role_name)

    # Set up Instance Profile.
    instance_profile_name = '-'.join(['role', config['PROJECT_NAME'], config['ENVIRONMENT']])
    instance_profile = iam_connection.create_instance_profile(instance_profile_name)
    instance_profile.name = instance_profile_name
    iam_connection.add_role_to_instance_profile(instance_profile_name, role_name)


    # Attach Inline Policies to Role.
    for inline_policy in inline_policies:
        role_policy_name = '-'.join(['policy', config['PROJECT_NAME'], config['ENVIRONMENT'], '{:08x}'.format(random.randrange(2**32))])
        iam_connection.put_role_policy(role_name, role_policy_name, inline_policy)

    # Allow time for Role to register with Amazon IAM service.
    time.sleep(5) # Required 5-second sleep.
    return instance_profile

def upload_ssl_certificate(public_key, private_key, certificate_chain=None, name=None):
    # Connect to the Amazon Identity and Access Management (Amazon IAM) service.
    iam_connection = connect_iam()

    # Generate Certificate name.
    if not name:
        name = '-'.join(['crt', config['PROJECT_NAME'], config['ENVIRONMENT']])

    # Read SSL Public Key.
    with open(public_key, 'r') as public_key_file:
        public_key = public_key_file.read()

    # Read SSL Private Key.
    with open(private_key, 'r') as private_key_file:
        private_key = private_key_file.read()

    # Read SSL Certificate Chain, if it was specified.
    if certificate_chain:
        with open(certificate_chain, 'r') as certificate_chain_file:
            certificate_chain = certificate_chain_file.read()

    # Delete Server Certificate, if one exists.
    try:
        logger.info('Deleting Server Certificate (%s).' % name)
        iam_connection.delete_server_cert(name)
        logger.info('Deleted Server Certificate (%s).' % name)
    except boto.exception.BotoServerError as error:
        if error.status == 400: # Bad Request
            logger.error('Couldn\'t delete Server Certificate (%s) due to an incompatible filename Error %s: %s.' % (name, error.status, error.reason))
        if error.status == 404: # Not Found
            logger.error('Couldn\'t delete Server Certificate (%s) due to Error %s: %s.' % (name, error.status, error.reason))

    # Upload the SSL Certificate to Amazon IAM.
    cert_arn = None
    try:
        logger.info('Uploading server certificate (%s).' % name)
        response = iam_connection.upload_server_cert(name, public_key, private_key, certificate_chain)
        logger.info('Uploaded server certificate (%s).' % name)
        server_certificate_id = response['upload_server_certificate_response']\
                                        ['upload_server_certificate_result']\
                                        ['server_certificate_metadata']\
                                        ['server_certificate_id']
        cert_arn = response['upload_server_certificate_response']\
                           ['upload_server_certificate_result']\
                           ['server_certificate_metadata']\
                           ['arn']
        time.sleep(5) # required 5-second sleep
    except boto.exception.BotoServerError as error:
        if error.status == 400: # Bad Request
            logger.error('Couldn\'t upload server certificate (%s) due to an issue with its contents and/or formatting Error %s: %s.' % (name, error.status, error.reason))
        if error.status == 409: # Conflict
            logger.error('Couldn\'t upload server certificate (%s) due to Error %s: %s.' % (name, error.status, error.reason))

    return cert_arn



#!/usr/bin/env python
from distutils.core import setup

setup(
    name = 'skyman',
    packages = ['sky'],
    version = '0.1',
    description = 'Sky is a tool that allows quick and repeatable deployment and management of cloud environments.',
    author = 'Jared Contrascere',
    author_email = 'jcontra@gmail.com',
    url = 'https://github.com/libretees/sky',

    install_requires=['boto==2.36.0'],
    entry_points={
        'console_scripts': [
            'sky = sky.main:main',
        ]
    },
    classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Unix',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Software Distribution',
          'Topic :: System :: Systems Administration',
    ],
)



import os
import sys
import logging

logger = logging.getLogger(__name__)

class ReadyObject(dict):

    def __init__(self):
        super(ReadyObject, self).__init__()

    def __getattr__(self, attr):
        try:
            value = self[attr]
        except KeyError as error:
            logger.error('%s is not available.')
        return value

    def __setattr__(self, key, value):
        self[key] = value

ready = ReadyObject()

CREATION_MODE = None
EPHEMERAL = 'ephemeral'
PERMANENT = 'permanent'

config = {
    'PROJECT_NAME':          None,
    'PROJECT_DIRECTORY':     None,
    'ENVIRONMENT':           None,
    'AWS_ACCOUNT_ID':        None,
    'AWS_ACCESS_KEY_ID':     None,
    'AWS_SECRET_ACCESS_KEY': None,
}


import sys
import time
import logging
import boto
from .compute import create_security_group
from .state import config

logger = logging.getLogger(__name__)

ENGINE = {
    'postgresql': 'postgres9.3',
    'mysql':      'MySQL5.6',
    'oracle':     'oracle-se1-11.2',
}

ENGINE_NAME = {
    'postgresql': 'postgres',
    'mysql':      'MySQL',
    'oracle':     'oracle-se1',
}

MAJOR_ENGINE_VERSION = {
    'postgresql': '9.3',
    'mysql':      '5.1.42',
    'oracle':     '11.2.0.2.v2',
}

INBOUND_PORT = {
    'postgresql': 5432,
    'mysql':      3306,
    'oracle':     1520,
}

def connect_rds():
    logger.debug('Connecting to the Amazon Relational Database Service (Amazon RDS).')
    rds = boto.connect_rds2(aws_access_key_id=config['AWS_ACCESS_KEY_ID'],
                            aws_secret_access_key=config['AWS_SECRET_ACCESS_KEY'])
    logger.debug('Connected to Amazon RDS.')
    
    return rds

def create_db_parameter_group(name=None, engine='postgresql'):
    # Connect to the Amazon Relational Database Service (Amazon RDS).
    rds_connection = connect_rds()

    # Affected by boto Issue #2677 : https://github.com/boto/boto/issues/2677


    # Generate Database Parameter Group name.
    if not name:
        name = '-'.join(['pg',
                         config['PROJECT_NAME'],
                         config['ENVIRONMENT'],])

    # Delete existing Database Parameter Group.
    try:
        rds_connection.delete_db_parameter_group(name)
    except boto.rds2.exceptions.DBParameterGroupNotFound as error:
        pass

    db_parameter_group = rds_connection.create_db_parameter_group(name,                                                    # db_parameter_group_name
                                                                  ENGINE[engine],                                          # db_parameter_group_family
                                                                  description=' '.join([config['PROJECT_NAME'], 'Parameter Group'])) # description

    # Construct Database Parameter Group ARN.
    region = 'us-east-1'
    db_parameter_group_arn = 'arn:aws:rds:%s:%s:pg:%s' % (region, config['AWS_ACCOUNT_ID'], name)

    # Tag Database Subnet Group.
    logger.debug('Tagging Amazon RDS Resource (%s).' % db_parameter_group_arn)
    rds_connection.add_tags_to_resource(db_parameter_group_arn,           # resource_name
                                        [('Name'       , name         ),  # tags
                                         ('Project'    , config['PROJECT_NAME'] ),
                                         ('Environment', config['ENVIRONMENT']  )])
    logger.debug('Tagged Amazon RDS Resource (%s).' % db_parameter_group_arn)

    return db_parameter_group

def create_db_subnet_group(subnets, name=None):
    # Connect to the Amazon Relational Database Service (Amazon RDS).
    rds_connection = connect_rds()

    # Generate Database Subnet Group name.
    if not name:
        name = '-'.join(['subgrp',
                         config['PROJECT_NAME'],
                         config['ENVIRONMENT'],])

    # Delte existing Database Subnet Group.
    try:
        rds_connection.delete_db_subnet_group(name)
    except boto.exception.JSONResponseError as error:
        if error.code == 'DBSubnetGroupNotFoundFault':
            pass

    # Create Database Subnet Group.
    subnet = None
    try:
        subnet = rds_connection.create_db_subnet_group(name,                                        #db_subnet_group_name
                                                       ' '.join([config['PROJECT_NAME'], 'DB Subnet Group']), #db_subnet_group_description
                                                       [subnet.id for subnet in subnets])           #subnet_ids
    except boto.rds2.exceptions.DBSubnetGroupAlreadyExists as error:
        if error.code == 'DBSubnetGroupAlreadyExists':
            subnet = rds_connection.describe_db_subnet_groups(name)

    # Construct Database Subnet Group ARN.
    region = 'us-east-1'
    db_subnet_group_arn = 'arn:aws:rds:%s:%s:subgrp:%s' % (region, config['AWS_ACCOUNT_ID'], name)

    # Tag Database Subnet Group.
    logger.debug('Tagging Amazon RDS Resource (%s).' % db_subnet_group_arn)
    rds_connection.add_tags_to_resource(db_subnet_group_arn,              # resource_name
                                        [('Name'       , name         ),  # tags
                                         ('Project'    , config['PROJECT_NAME'] ),
                                         ('Environment', config['ENVIRONMENT']  )])
    logger.debug('Tagged Amazon RDS Resource (%s).' % db_subnet_group_arn)

    return subnet

def create_option_group(name=None, engine='postgresql'):
    # Connect to the Amazon Relational Database Service (Amazon RDS).
    rds_connection = connect_rds()

    # Generate Option Group name.
    if not name:
        name = '-'.join(['og',
                         config['PROJECT_NAME'],
                         config['ENVIRONMENT'],])

    # Delete Option Group.
    try:
        rds_connection.delete_option_group(name)
    except boto.exception.JSONResponseError as error:
        if error.status == 404 and error.reason == 'Not Found' and error.body['Error']['Code'] == 'OptionGroupNotFoundFault':
            pass
        else:
            raise boto.exception.JSONResponseError(error.status, error.reason, body=error.body)

    # Create Option Group.
    option_group = rds_connection.create_option_group(name,                                     # option_group_name
                                                      ENGINE_NAME[engine],                      # engine_name
                                                      MAJOR_ENGINE_VERSION[engine],             # major_engine_version
                                                      ' '.join([config['PROJECT_NAME'], 'Option Group']), # option_group_description
                                                      tags=None)

    # Construct Option Group ARN.
    region = 'us-east-1'
    option_group_arn = 'arn:aws:rds:%s:%s:og:%s' % (region, config['AWS_ACCOUNT_ID'], name)

    # Tag Option Group.
    logger.debug('Tagging Amazon RDS Resource (%s).' % option_group_arn)
    rds_connection.add_tags_to_resource(option_group_arn   ,              # resource_name
                                        [('Name'       , name         ),  # tags
                                         ('Project'    , config['PROJECT_NAME'] ),
                                         ('Environment', config['ENVIRONMENT']  )])
    logger.debug('Tagged Amazon RDS Resource (%s).' % option_group_arn)

    return option_group

def create_database(vpc, subnets, name=None, engine='postgresql', application_instances=None, security_groups=None, publicly_accessible=False, multi_az=False, db_parameter_group=None, option_group=None):
    # Connect to the Amazon Relational Database Service (Amazon RDS).
    rds_connection = connect_rds()

    db_subnet_group = create_db_subnet_group(subnets)

    if not name:
        name = '-'.join(['db',
                         config['PROJECT_NAME'],
                         config['ENVIRONMENT'],])

    if not db_parameter_group:
        db_parameter_group = create_db_parameter_group(engine=engine)

    if not option_group:
        option_group = create_option_group(engine=engine)

    db_parameter_group_name = db_parameter_group['CreateDBParameterGroupResponse']\
                                                ['CreateDBParameterGroupResult']\
                                                ['DBParameterGroup']\
                                                ['DBParameterGroupName']

    db_subnet_group_name = db_subnet_group['CreateDBSubnetGroupResponse']\
                                          ['CreateDBSubnetGroupResult']\
                                          ['DBSubnetGroup']\
                                          ['DBSubnetGroupName']

    option_group_name = option_group['CreateOptionGroupResponse']\
                                    ['CreateOptionGroupResult']\
                                    ['OptionGroup']\
                                    ['OptionGroupName']

    if not security_groups:
        if application_instances:
            # Create rule(s) allowing traffic from application server security group(s).
            application_security_group_ids = set([group.id for instance in application_instances for group in instance.groups])
            inbound_rules = list()
            for application_security_group_id in application_security_group_ids:
                inbound_rule = ('TCP:' + str(INBOUND_PORT[engine]), application_security_group_id)
                inbound_rules.append(inbound_rule)

        sg_name = '-'.join(['gp', config['PROJECT_NAME'], config['ENVIRONMENT'], 'db'])
        security_groups = [create_security_group(vpc, name=sg_name
                                                    , allowed_inbound_traffic=inbound_rules if application_instances else None
                                                    , allowed_outbound_traffic=[('HTTP',  '0.0.0.0/0')
                                                                               ,('HTTPS', '0.0.0.0/0')
                                                                               ,('DNS',   '0.0.0.0/0')])]

    db_instance = rds_connection.create_db_instance(name,                                                     # db_instance_identifier
                                                    5,                                                        # allocated_storage
                                                    'db.t2.micro',                                            # db_instance_class
                                                    ENGINE_NAME[engine],                                      # engine
                                                    'username',                                               # master_username
                                                    'password',                                               # master_user_password
                                                    db_name=None,
                                                    db_security_groups=None,                                  # Used in EC2-Classic.
                                                    vpc_security_group_ids=[sg.id for sg in security_groups],
                                                    availability_zone=None,                                   # Used in EC2-Classic.
                                                    db_subnet_group_name=db_subnet_group_name,                # Required for EC2-VPC Database Instances.
                                                    preferred_maintenance_window=None,
                                                    db_parameter_group_name=db_parameter_group_name,
                                                    backup_retention_period=None,
                                                    preferred_backup_window=None,
                                                    port=None,
                                                    multi_az=multi_az,
                                                    engine_version=None,
                                                    auto_minor_version_upgrade=None,
                                                    license_model=None,
                                                    iops=None,
                                                    option_group_name=option_group_name,
                                                    character_set_name=None,
                                                    publicly_accessible=publicly_accessible,
                                                    tags=None)

    # Construct Database Instance ARN.
    region = 'us-east-1'
    database_arn = 'arn:aws:rds:%s:%s:db:%s' % (region, config['AWS_ACCOUNT_ID'], name)

    # Tag Database Instance.
    logger.debug('Tagging Amazon RDS Resource (%s).' % database_arn)
    rds_connection.add_tags_to_resource(database_arn,                     # resource_name
                                        [('Name'       , name         ),  # tags
                                         ('Project'    , config['PROJECT_NAME'] ),
                                         ('Environment', config['ENVIRONMENT']  )])
    logger.debug('Tagged Amazon RDS Resource (%s).' % database_arn)

    # Get Database Endpoint.
    logger.info('Getting endpoint for database (%s).' % name)
    endpoint = None
    while not endpoint:
        response = rds_connection.describe_db_instances(db_instance_identifier=name,
                                                        filters=None,
                                                        max_records=None,
                                                        marker=None)
        endpoint = response['DescribeDBInstancesResponse']\
                           ['DescribeDBInstancesResult']\
                           ['DBInstances'][-1]\
                           ['Endpoint']
        if not endpoint:
            logger.debug('Waiting for database endpoint...')
            time.sleep(1)
        else:
            logger.info('Got database endpoint (%s).' % endpoint)
    db_instance['endpoint'] = endpoint

    return db_instance



import sys
import time
import re
import ipaddress
import logging
from operator import itemgetter
import boto
from .state import config

logger = logging.getLogger(__name__)

def connect_vpc():
    logger.debug('Connecting to the Amazon Virtual Private Cloud (Amazon VPC) service.')
    vpc = boto.connect_vpc(aws_access_key_id=config['AWS_ACCESS_KEY_ID'],
                           aws_secret_access_key=config['AWS_SECRET_ACCESS_KEY'])
    logger.debug('Connected to Amazon VPC.')

    return vpc

def validate_cidr_block(cidr_block):
    try:
        logger.debug('Validating CIDR block (%s).' % cidr_block)

        # Split CIDR block into Network IP and Netmask components.
        network_ip, netmask = itemgetter(0, 1)(cidr_block.split('/'))
        netmask = int(netmask)

        # Ensure that CIDR block conforms to RFC 1918.
        assert (re.search(r'^10\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip) or \
                re.search(r'^172\.(1[6-9]|2[0-9]|3[0-1])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip) or \
                re.search(r'^192\.168\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip)) and \
               (re.search(r'(^[1-2]?[0-9]$)|(^3[0-2]$)', str(netmask)))

        # Validate netmask.
        if re.search(r'^10\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip):
            assert netmask >= 8
            logger.debug('Valid Class A Private Network CIDR block given (%s).' % cidr_block)
        elif re.search(r'^172\.(1[6-9]|2[0-9]|3[0-1|)\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip):
            assert netmask >= 12
            logger.debug('Valid Class B Private Network CIDR block given (%s).' % cidr_block)
        elif re.search(r'^192\.168\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])\.([0-9]|[1-9][0-9]|[1-2][0-5][0-5])$', network_ip):
            assert netmask >= 16
            logger.debug('Valid Class C Private Network CIDR block given (%s).' % cidr_block)

        # Ensure that netmask is compatible with Amazon VPC.
        if not (netmask >= 16 and netmask <= 28):
            logger.error('Amazon VPC service requires CIDR block sizes between a /16 netmask and /28 netmask.')
            assert False

        logger.debug('CIDR block validated (%s).' % cidr_block)
        return True
    except AssertionError as error:
        logger.error('Invalid CIDR block given (%s).' % cidr_block)
        return False

def create_network(name=None, internet_connected=False, **kwargs):
    # Deferred import to resolve interdependency between .networking and .compute modules.
    from .compute import connect_ec2

    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    if 'cidr_block' in kwargs:
        cidr_block = kwargs['cidr_block']
    elif 'network_class' in kwargs:
        network_class = kwargs['network_class']
        cidr_block = '10.0.0.0/16' if network_class.upper() == 'A' \
                else '172.16.0.0/16' if network_class.upper() == 'B' \
                else '192.168.0.0/16' if network_class.upper() == 'C' \
                else '0.0.0.0/0'
    else:
        raise TypeError( "Value for cidr_block or network_class arguments must be specified." )
        
    if not validate_cidr_block(cidr_block):
        sys.exit(1)

    # Generate Virtual Private Cloud (VPC) name, if needed.
    if not name:
        name = '-'.join(['vpc', config['PROJECT_NAME'], config['ENVIRONMENT']])

    # Create Virtual Private Cloud (VPC).
    network = None
    try:
        logger.info('Creating Virtual Private Cloud (VPC) (%s) with CIDR block (%s).' % (name, cidr_block))
        network = vpc_connection.create_vpc(cidr_block,                 # cidr_block
                                            instance_tenancy='default',
                                            dry_run=False)
        logger.info('Created Virtual Private Cloud (VPC) (%s).' % name)
    except boto.exception.EC2ResponseError as error:
        if error.status == 400: # Bad Request
            logger.error('Error %s: %s. Could not create VPC (%s). %s' % (error.status, error.reason, name, error.message))

    # Tag Virtual Private Cloud (VPC).
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([network.id], {'Name': name,
                                                               'Project': config['PROJECT_NAME'],
                                                               'Environment': config['ENVIRONMENT'],})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidVpcID.NotFound': # VPC hasn't registered with Virtual Private Cloud (VPC) service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    # Tag default Security Group.
    security_groups = ec2_connection.get_all_security_groups(filters={'vpc-id': network.id,})
    for security_group in security_groups:
        tagged = False
        security_group_name = '-'.join(['gp', config['PROJECT_NAME'], config['ENVIRONMENT'], 'default'])
        while not tagged:
            try:
                tagged = ec2_connection.create_tags([security_group.id], {'Name': security_group_name,
                                                                          'Project': config['PROJECT_NAME'],
                                                                          'Environment': config['ENVIRONMENT'],
                                                                          'Type': 'default',})
            except boto.exception.EC2ResponseError as error:
                if error.code == 'InvalidID': # Security Group hasn't registered with Virtual Private Cloud (VPC) service yet.
                    pass
                else:
                    raise boto.exception.EC2ResponseError

    # Tag Main Route Table.
    route_tables = vpc_connection.get_all_route_tables(filters={'vpc-id': network.id,})
    for route_table in route_tables:
        tagged = False
        route_table_name = '-'.join(['rtb', config['PROJECT_NAME'], config['ENVIRONMENT'], 'main'])
        while not tagged:
            try:
                tagged = ec2_connection.create_tags([route_table.id], {'Name': route_table_name,
                                                                       'Project': config['PROJECT_NAME'],
                                                                       'Environment': config['ENVIRONMENT'],
                                                                       'Type': 'main',})
            except boto.exception.EC2ResponseError as error:
                if error.code == 'InvalidID': # Route Table hasn't registered with Virtual Private Cloud (VPC) service yet.
                    pass
                else:
                    raise boto.exception.EC2ResponseError

    # Tag Access Control Lists (ACLs).
    acls = vpc_connection.get_all_network_acls(filters={'vpc-id': network.id,})
    for acl in acls:
        tagged = False
        acl_name = '-'.join(['acl', config['PROJECT_NAME'], config['ENVIRONMENT']])
        while not tagged:
            try:
                tagged = ec2_connection.create_tags([acl.id], {'Name': acl_name,
                                                               'Project': config['PROJECT_NAME'],
                                                               'Environment': config['ENVIRONMENT'],})
            except boto.exception.EC2ResponseError as error:
                if error.code == 'InvalidNetworkAclID.NotFound': # ACL hasn't registered with Virtual Private Cloud (VPC) service yet.
                    pass
                else:
                    raise boto.exception.EC2ResponseError

    # Tag DHCP Options Set.
    dhcp_options = vpc_connection.get_all_dhcp_options(network.dhcp_options_id)
    for dhcp_option in dhcp_options:
        tagged = False
        dhcp_option_name = '-'.join(['dopt', config['PROJECT_NAME'], config['ENVIRONMENT']])
        while not tagged:
            try:
                tagged = ec2_connection.create_tags([dhcp_option.id], {'Name': dhcp_option_name,
                                                                       'Project': config['PROJECT_NAME'],
                                                                       'Environment': config['ENVIRONMENT'],})
            except boto.exception.EC2ResponseError as error:
                if error.code == 'InvalidID': # DHCP Options Set hasn't registered with Virtual Private Cloud (VPC) service yet.
                    pass
                else:
                    raise boto.exception.EC2ResponseError

    if internet_connected:
        attach_internet_gateway(network)

    return network

def attach_internet_gateway(vpc):
    # Deferred import to resolve interdependency between .networking and .compute modules.
    from .compute import connect_ec2

    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Create Internet Gateway.
    internet_gateway = vpc_connection.create_internet_gateway(dry_run=False)

    # Tag Internet Gateway.
    tagged = False
    internet_gateway_name = '-'.join(['igw', config['PROJECT_NAME'], config['ENVIRONMENT']])
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([internet_gateway.id], {'Name': internet_gateway_name,
                                                                        'Project': config['PROJECT_NAME'],
                                                                        'Environment': config['ENVIRONMENT'],})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidInternetGatewayID.NotFound': # IGW hasn't registered with Virtual Private Cloud (VPC) service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    # Get name of VPC.
    vpc_tags = ec2_connection.get_all_tags(filters={'resource-id': vpc.id,
                                                    'resource-type': 'vpc'})
    vpc_name = [tag.value for tag in vpc_tags if tag.name == 'Name'][0]

    # Attach Internet Gateway to Virtual Private Cloud (VPC).
    logger.info('Attaching Internet Gateway (%s) to VPC (%s).' % (internet_gateway_name, vpc_name))
    attached = vpc_connection.attach_internet_gateway(internet_gateway.id, # internet_gateway_id
                                                      vpc.id,              # vpc_id
                                                      dry_run=False)
    if attached:
        logger.info('Attached Internet Gateway (%s).' % internet_gateway_name)
    else:
        logger.error('Could not attach Internet Gateway (%s).' % internet_gateway_name)

    return internet_gateway

def create_route_table(vpc, name=None, internet_access=False):
    # Deferred import to resolve interdependency between .networking and .compute modules.
    from .compute import connect_ec2
    
    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Create Route Table.
    route_table = vpc_connection.create_route_table(vpc.id)

    if internet_access:
        # Get an Internet Gateway associated to the VPC.
        internet_gateways = vpc_connection.get_all_internet_gateways(filters={'attachment.vpc-id': vpc.id,
                                                                              'attachment.state': 'available',})
        if internet_gateways:
            internet_gateway = [internet_gateway for internet_gateway in internet_gateways][0] if len(internet_gateways) else None
        else:
            internet_gateway = attach_internet_gateway(vpc)

        # Add route to Internet to Route Table.
        vpc_connection.create_route(route_table.id,    # route_table_id
                                    '0.0.0.0/0',       # destination_cidr_block
                                    gateway_id=internet_gateway.id,
                                    instance_id=None,
                                    interface_id=None,
                                    vpc_peering_connection_id=None,
                                    dry_run=False)

        # Refresh Route Table.
        refreshed = False
        while not refreshed:
            try:
                route_table = vpc_connection.get_all_route_tables(route_table.id)
                route_table = route_table[0] if len(route_table) else None
                refreshed = True
            except boto.exception.EC2ResponseError as error:
                if error.code == 'InvalidRouteTableID.NotFound':
                    pass

    # Generate Route Table name.
    route_tables = vpc_connection.get_all_route_tables(filters={'vpc-id': vpc.id,})
    public_route_tables = vpc_connection.get_all_route_tables(filters={'vpc-id': vpc.id,
                                                                       'route.destination-cidr-block': '0.0.0.0/0'})
    suffix = str(len(public_route_tables) if internet_access else len(route_tables)-len(public_route_tables)-1)

    if not name:
        name = '-'.join(['rtb', \
                         config['PROJECT_NAME'], \
                         config['ENVIRONMENT'], \
                         'public' if internet_access else 'private', \
                         suffix])
    route_table.name = name

    # Tag Route Table.
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([route_table.id], {'Name': route_table.name,
                                                                   'Project': config['PROJECT_NAME'],
                                                                   'Environment': config['ENVIRONMENT'],
                                                                   'Type': 'public' if internet_access else 'private',})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidRouteTableID.NotFound': # Route Table hasn't registered with Virtual Private Cloud (VPC) service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError(error.status, error.reason)

    logger.info('Created Route Table (%s).' % route_table.name)
    return route_table

def create_subnets(vpc, zones='All', count=1, byte_aligned=False, balanced=False, public=False):
    # Deferred import to resolve interdependency between .networking and .compute modules.
    from .compute import connect_ec2
    
    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Break CIDR block into IP and Netmask components.
    network_ip, netmask = get_cidr_block_components(vpc.cidr_block)

    # Get/Validate Availability Zones associated with the current region.
    if isinstance(zones, str) and zones.lower() == 'all':
        zones = None
    elif isinstance(zones, str):
        zones = [zone.strip() for zone in zones.lower().split(',')]
    zones = ec2_connection.get_all_zones(zones)

    # Get the number of Subnets in each zone, so that a Subnet name can be computed.
    for zone in zones:
        offset = len(vpc_connection.get_all_subnets(filters={'vpc-id': vpc.id,
                                                             'availability-zone': zone.name,
                                                             'tag:Type': 'public' if public else 'private',}))
        zone.offset = offset

    # Get the number of Subnets within the specified VPC.
    num_subnets = len(vpc_connection.get_all_subnets(filters={'vpc-id': vpc.id,}))

    # Calculate Subnet netmask.
    subnet_netmask = netmask+len(bin(num_subnets+len(zones)*count))-3

    if balanced:
        # Balance between network expandability and network size.
        subnet_netmask = subnet_netmask+(28-subnet_netmask)//2 if subnet_netmask < 28 else subnet_netmask

    if byte_aligned:
        # Align CIDR block to nearest byte, if possible.
        subnet_netmask = subnet_netmask+8-(subnet_netmask%8) if subnet_netmask < 24 else subnet_netmask

    # Create Route Table for Public/Private Subnets.
    route_table = create_route_table(vpc, internet_access=True) if public else create_route_table(vpc, internet_access=False)

    # Create Subnets.
    subnets = list()
    for i, zone in enumerate(sorted(zones*count, key=lambda zone: zone.name)):
        # Generate Subnet name.
        suffix = '-' + str(1+zone.offset+(i%count)).zfill(len(str(zone.offset+count)))
        subnet_name = '-'.join(['subnet', \
                                config['PROJECT_NAME'], \
                                config['ENVIRONMENT'], \
                                zone.name, \
                                ('public' if public else 'private') + suffix])

        # Shorten Subnet name, if needed and if possible.
        shortened = dict(dict.fromkeys(['resource_type', 'environment', 'subnet_type'], False))
        while len(subnet_name) > 37:
            if not shortened['resource_type']:
                subnet_name = subnet_name.replace('subnet-', '')
                shortened['resource_type'] = True
                continue
            if not shortened['subnet_type']:
                subnet_name = subnet_name.replace('public', 'pub') if public else subnet_name.replace('private', 'priv')
                shortened['subnet_type'] = True
                continue
            if not shortened['environment']:
                if config['ENVIRONMENT'] == 'prod':
                    subnet_name = subnet_name.replace('prod', 'prd')
                elif config['ENVIRONMENT'] == 'staging':
                    subnet_name = subnet_name.replace('staging', 'stg')
                shortened['environment'] = True
                continue
            break # Give up.

        # Create Subnet CIDR block.
        subnet_network_ip = (network_ip | (num_subnets+i << 32-subnet_netmask))
        subnet_cidr_block = str(subnet_network_ip >> 24) + '.' + \
                            str((subnet_network_ip >> 16) & 255) + '.' + \
                            str((subnet_network_ip >> 8) & 255) + '.' + \
                            str(subnet_network_ip & 255) + '/' + \
                            str(subnet_netmask)

        # Create Subnet.
        subnet = create_subnet(vpc, zone, subnet_cidr_block, subnet_name, route_table)

        # Add Subnet to list.
        if subnet:
            subnets.append(subnet)

    return subnets

def create_subnet(vpc, zone, cidr_block, subnet_name=None, route_table=None):
    # Deferred import to resolve interdependency between .networking and .compute modules.
    from .compute import connect_ec2
    
    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Break CIDR block into IP and Netmask components.
    network_ip, netmask = get_cidr_block_components(cidr_block)

    # Create Subnet.
    try:
        logger.info('Creating Subnet (%s).' % subnet_name)
        subnet = vpc_connection.create_subnet(vpc.id,                      # vpc_id
                                              cidr_block,                  # cidr_block
                                              availability_zone=zone.name,
                                              dry_run=False)
        logger.info('Created Subnet (%s) with %s available IP addresses.' % (subnet_name, '{:,}'.format(get_network_capacity(netmask))))
    except boto.exception.BotoServerError as error:
        if error.status == 400: # Bad Request
            logger.error('Error %s: %s. Couldn\'t create Subnet (%s).' % (error.status, error.reason, subnet_name))
            if error.code == 'SubnetLimitExceeded':
                num_subnets = len(vpc_connection.get_all_subnets(filters={'vpc-id':vpc.id}))
                logging.error('%d Subnets exist within the VPC' % num_subnets)
                logging.error('Refer to the VPC User Guide for Amazon VPC Limits.')
                sys.exit(1)

    # Associate Subnet to Route Table.
    public = False
    if route_table:
        association_id = vpc_connection.associate_route_table(route_table.id, # route_table_id
                                                              subnet.id,      # subnet_id
                                                              dry_run=False)
        if len(association_id):
            logger.debug('Subnet (%s) associated to (%s).' % (subnet_name, route_table.id))
        else:
            logger.error('Subnet (%s) not associated to (%s).' % (subnet_name, route_table.id))

        # Determine Subnet type.
        public = [route for route in route_table.routes if route.gateway_id and route.destination_cidr_block == '0.0.0.0/0']

    # Tag Subnet.
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([subnet.id], {'Name': subnet_name,
                                                              'Project': config['PROJECT_NAME'],
                                                              'Environment': config['ENVIRONMENT'],
                                                              'Type': 'public' if public else 'private',})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidSubnetID.NotFound': # Subnet hasn't registered with Virtual Private Cloud (VPC) service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    return subnet

def get_network_capacity(netmask):
    # Calculate the number of available IP addresses on a given network.
    available_ips = (0xffffffff ^ (0xffffffff << 32-int(netmask) & 0xffffffff))-4 # 4 addresses are reserved by Amazon
                                                                                  # for IP networking purposes.
                                                                                  # .0 for the network, .1 for the gateway,
                                                                                  # .3 for DHCP services, and .255 for broadcast.
    return available_ips

def get_default_vpc():
    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Get all Virtual Private Clouds (VPCs).
    vpcs = vpc_connection.get_all_vpcs()

    # Return default VPC.
    default_vpc = [vpc for vpc in vpcs if vpc.is_default]
    return default_vpc[0] if len(default_vpc) else None

def get_cidr_block_components(cidr_block):
    # Break CIDR block into IP and Netmask components.
    network_ip, netmask = itemgetter(0, 1)(cidr_block.split('/'))
    network_ip = int(ipaddress.IPv4Address(network_ip))
    netmask = int(netmask)

    return (network_ip, netmask)



import re
import time
import random
import logging
from operator import itemgetter
import boto
from .networking import connect_vpc, create_route_table
from .state import config

logger = logging.getLogger(__name__)

def connect_ec2():
    logger.debug('Connecting to the Amazon Elastic Compute Cloud (Amazon EC2) service.')
    ec2 = boto.connect_ec2(aws_access_key_id=config['AWS_ACCESS_KEY_ID'],
                           aws_secret_access_key=config['AWS_SECRET_ACCESS_KEY'])
    logger.debug('Connected to Amazon EC2.')

    return ec2

def create_security_group(vpc, name=None, allowed_inbound_traffic=None, allowed_outbound_traffic=None):
    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Generate Security Group name.
    if not name:
        name = '-'.join(['gp', config['PROJECT_NAME'], config['ENVIRONMENT']])

    if not allowed_inbound_traffic:
        allowed_inbound_traffic = []

    if not allowed_outbound_traffic:
        allowed_outbound_traffic = []

    # Create Security Group.
    logger.info('Creating Security Group (%s).' % name)
    security_group = ec2_connection.create_security_group(name, 'Security Group Description', vpc_id=vpc.id)
    logger.info('Created Security Group (%s).' % name)

    # Set up inbound/outbound rules.
    ec2_connection.revoke_security_group_egress(security_group.id, -1, from_port=0, to_port=65535, cidr_ip='0.0.0.0/0')
    for (protocol, target, rule_type) in [(traffic[0].upper(), traffic[1], 'inbound') for traffic in allowed_inbound_traffic] + \
                                         [(traffic[0].upper(), traffic[1], 'outbound') for traffic in allowed_outbound_traffic]:

        # Determine whether target is a CIDR block or a Security Group.
        is_cidr_ip = re.search('^(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.'+      # A in A.B.C.D
                               '(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.' +      # B in A.B.C.D
                               '(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.' +      # C in A.B.C.D
                               '(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)'   +      # D in A.B.C.D
                               '(/(([1-2]?[0-9])|(3[0-2])))$', str(target)) # /0 through /32
        target_cidr_ip = target if is_cidr_ip else None
        target_group = ec2_connection.get_all_security_groups(group_ids=[target])[-1] if not is_cidr_ip else None

        # Determine port range for TCP and UDP rules.
        port = None
        if protocol[:3] in ['TCP', 'UDP']:
            protocol, port = itemgetter(0, 1)(protocol.split(':'))
            if re.search(r'^\d+\-\d+$', port):
                from_port, to_port = itemgetter(0, 1)(port.split('-'))
            else:
                from_port, to_port = port, port

        # Create inbound rules.
        if rule_type == 'inbound':
            if protocol == 'HTTP':
                security_group.authorize(ip_protocol='tcp', from_port=80, to_port=80, src_group=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'HTTPS':
                security_group.authorize(ip_protocol='tcp', from_port=443, to_port=443, src_group=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'TCP':
                security_group.authorize(ip_protocol='tcp', from_port=int(from_port), to_port=int(to_port), src_group=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'UDP':
                security_group.authorize(ip_protocol='udp', from_port=int(from_port), to_port=int(to_port), src_group=target_group, cidr_ip=target_cidr_ip)
            logger.info('Security Group (%s) allowed inbound %s traffic from %s.' % (name, protocol + (' Port %s' % port if port else ''), target))

        # Create outbound rules.
        if rule_type == 'outbound':
            if protocol == 'HTTP':
                ec2_connection.authorize_security_group_egress(security_group.id, 'tcp', from_port=80, to_port=80, src_group_id=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'HTTPS':
                ec2_connection.authorize_security_group_egress(security_group.id, 'tcp', from_port=443, to_port=443, src_group_id=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'DNS':
                ec2_connection.authorize_security_group_egress(security_group.id, 'tcp', from_port=53, to_port=53, src_group_id=target_group, cidr_ip=target_cidr_ip)
                ec2_connection.authorize_security_group_egress(security_group.id, 'udp', from_port=53, to_port=53, src_group_id=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'TCP':
                 ec2_connection.authorize_security_group_egress(security_group.id, 'tcp', from_port=int(from_port), to_port=int(to_port), src_group_id=target_group, cidr_ip=target_cidr_ip)
            elif protocol == 'UDP':
                 ec2_connection.authorize_security_group_egress(security_group.id, 'udp', from_port=int(from_port), to_port=int(to_port), src_group_id=target_group, cidr_ip=target_cidr_ip)
            logger.info('Security Group (%s) allowed outbound %s traffic to %s.' % (name, protocol + (' Port %s' % port if port else ''), target))

    # Tag Security Group.
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags(security_group.id, {'Name': name,
                                                                    'Project': config['PROJECT_NAME'],
                                                                    'Environment': config['ENVIRONMENT'],})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidID': # Security Group hasn't registered with EC2 service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    return security_group

def create_load_balancer(vpc, subnets, name=None, security_groups=None, ssl_certificate=None):
    # Connect to the Amazon EC2 Load Balancing (Amazon ELB) service.
    logger.debug('Connecting to the Amazon EC2 Load Balancing (Amazon ELB) service.')
    elb_connection = boto.connect_elb()
    logger.debug('Connected to the Amazon EC2 Load Balancing (Amazon ELB) service.')

    # Set up default security group, if necessary
    if not security_groups:
        security_group_name = '-'.join(['gp', config['PROJECT_NAME'], config['ENVIRONMENT'], 'elb'])
        security_groups = [create_security_group(vpc,
                                                 name=security_group_name,
                                                 allowed_inbound_traffic=[('HTTP',   '0.0.0.0/0')
                                                                         ,('HTTPS',  '0.0.0.0/0')],
                                                 allowed_outbound_traffic=[('HTTP',  '0.0.0.0/0')
                                                                          ,('HTTPS', '0.0.0.0/0')
                                                                          ,('DNS',   '0.0.0.0/0')])]

    # Generate Elastic Load Balancer (ELB) name.
    if not name:
        name = '-'.join(['elb', config['PROJECT_NAME'], config['ENVIRONMENT']])

    # Delete existing Elastic Load Balancer (ELB).
    logger.info('Deleting Elastic Load Balancer (%s).' % name)
    try:
        elb_connection.delete_load_balancer(name)
    except boto.exception.BotoServerError as error:
        if error.status == 400: # Bad Request
            logger.error('Couldn\'t delete Elastic Load Balancer (%s) due to a malformed request %s: %s.' % (name, error.status, error.reason))
        if error.status == 404: # Not Found
            logger.error('Elastic Load Balancer (%s) was not found. Error %s: %s.' % (name, error.status, error.reason))
    logger.info('Deleted Elastic Load Balancer (%s).' % name)

    # Create Security Group.
    if not security_groups:
        security_groups = create_security_group(vpc)

    listeners = [(80, 80, 'HTTP')]
    if ssl_certificate:
        listeners.append((443, 443, 'HTTPS', ssl_certificate))

    # Create Elastic Load Balancer (ELB).
    logger.info('Creating Elastic Load Balancer (%s).' % name)
    load_balancer = elb_connection.create_load_balancer(name, # name
                                                        None,     # zones         - Valid only for load balancers in EC2-Classic.
                                                        listeners=listeners,
                                                        subnets=[subnet.id for subnet in subnets],
                                                        security_groups=[security_group.id for security_group in security_groups],
                                                        scheme='internet-facing', # Valid only for load balancers in EC2-VPC.
                                                        complex_listeners=None)
    logger.info('Created Elastic Load Balancer (%s).' % name)

    return load_balancer

def create_nat_instances(vpc, public_subnets, private_subnets, security_groups=None, image_id=None):

    if not len(public_subnets) == len(private_subnets):
        raise RuntimeError('The number of Public/Private Subnets must match (Public: %d Private: %d).' % (len(public_subnets), len(private_subnets)))

    subnet_pairs = list(zip(sorted(public_subnets, key=lambda x: x.availability_zone), sorted(private_subnets, key=lambda x: x.availability_zone)))

    # Create NAT instances.
    nat_instances = list()
    for (public_subnet, private_subnet) in subnet_pairs:
        nat_instance = create_nat_instance(vpc, public_subnet, private_subnet, name=None, security_groups=None, image_id=None)
        nat_instances.append(nat_instance)

    return nat_instances

def create_nat_instance(vpc, public_subnet, private_subnet, name=None, security_groups=None, image_id=None):
    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Connect to the Amazon Virtual Private Cloud (Amazon VPC) service.
    vpc_connection = connect_vpc()

    # Create Security Group, if one was not specified.
    if not security_groups:
        sg_name = '-'.join(['gp', config['PROJECT_NAME'], config['ENVIRONMENT'], public_subnet.availability_zone, 'nat'])
        security_groups = [create_security_group(vpc, name=sg_name
                                                    , allowed_inbound_traffic=[('HTTP',   private_subnet.cidr_block)
                                                                              ,('HTTPS',  private_subnet.cidr_block)]
                                                    , allowed_outbound_traffic=[('HTTP',  '0.0.0.0/0')
                                                                               ,('HTTPS', '0.0.0.0/0')
                                                                               ,('DNS',   '0.0.0.0/0')])]

    # Get Amazon Linux VPC NAT AMI, if one was not specified.
    if not image_id:
        image_id = get_nat_image()

    # Generate name, if one was not specified.
    if not name:
        name = '-'.join(['ec2', config['PROJECT_NAME'], config['ENVIRONMENT'], public_subnet.availability_zone, 'nat'])

    # Create NAT Instance.
    nat_instance = create_instance(public_subnet, name=name, role='nat', security_groups=security_groups, image_id=image_id.id, internet_addressable=True)[0]

    # Disable source/destination checking.
    ec2_connection.modify_instance_attribute(nat_instance.id, attribute='sourceDestCheck', value=False, dry_run=False)

    # Create Route table.
    route_table_name = '-'.join(['rtb', config['PROJECT_NAME'], config['ENVIRONMENT'], public_subnet.availability_zone, 'private'])
    route_table = create_route_table(vpc, name=route_table_name, internet_access=False)

    # Wait for NAT instance to run.
    while (nat_instance.state == 'pending'):
        logger.debug('Waiting for NAT instance to run (%s)...' % nat_instance.id)
        time.sleep(1)
        nat_instance.update()

    # Add route to NAT Instance to Route Table.
    vpc_connection.create_route(route_table.id,    # route_table_id
                                '0.0.0.0/0',       # destination_cidr_block
                                gateway_id=None,
                                instance_id=nat_instance.id,
                                interface_id=None,
                                vpc_peering_connection_id=None,
                                dry_run=False)

    # Check for existing Route Table association.
    route_tables = vpc_connection.get_all_route_tables(filters={'vpc-id': vpc.id,})
    existing_association = [association.id for route_table in route_tables for association in route_table.associations if association.subnet_id == private_subnet.id]
    existing_association = existing_association[0] if existing_association else None

    # Associate Private Subnet to Route Table.
    if existing_association:
        association = vpc_connection.replace_route_table_association_with_assoc(existing_association, # association_id
                                                                                route_table.id,       # route_table_id
                                                                                dry_run=False)
    else:
        association = vpc_connection.associate_route_table(route_table.id,    # route_table_id
                                                           private_subnet.id, # subnet_id
                                                           dry_run=False)
    if len(association):
        logger.debug('Subnet (%s) associated to (%s).' % (private_subnet.id, route_table.name))
    else:
        logger.error('Subnet (%s) not associated to (%s).' % (private_subnet.id, route_table.name))

    # Clean up unused/orphaned Route Tables.
    route_tables = vpc_connection.get_all_route_tables(filters={'vpc-id': vpc.id,})
    main_route_table = vpc_connection.get_all_route_tables(filters={'vpc-id': vpc.id,
                                                                    'association.main': 'true'})[0] # Affected by boto Issue #1742 : https://github.com/boto/boto/issues/1742
    empty_route_tables = [route_table for route_table in route_tables if not len(route_table.associations) and not route_table.id == main_route_table.id]
    for route_table in empty_route_tables:
        try:
            vpc_connection.delete_route_table(route_table.id, dry_run=False)
        except boto.exception.EC2ResponseError as error:
            if error.code == 'DependencyViolation': # Route Table was not actually empty.
                pass

    return nat_instance

def create_instances(vpc, subnets, role=None, security_groups=None, script=None, instance_profile=None, os='ubuntu', image_id=None, internet_addressable=False):
    # Create a security group, if a security group was not specified.
    if not security_groups:
        security_groups = [create_security_group(vpc, allowed_inbound_traffic=[('HTTP',   '0.0.0.0/0')
                                                                              ,('HTTPS',  '0.0.0.0/0')]
                                                    , allowed_outbound_traffic=[('HTTP',  '0.0.0.0/0')
                                                                               ,('HTTPS', '0.0.0.0/0')
                                                                               ,('DNS',   '0.0.0.0/0')])]

    # Create EC2 instances.
    instances = list()
    for subnet in subnets:
        instance = create_instance(subnet, role=role, security_groups=security_groups, script=script, instance_profile=instance_profile, os=os, image_id=image_id, internet_addressable=internet_addressable)
        instances = instances + instance
    return instances

def create_instance(subnet, name=None, role=None, security_groups=None, script=None, instance_profile=None, os='ubuntu', image_id=None, internet_addressable=False):
    # Set up dictionary of OSes and their associated quick-start Amazon Machine Images (AMIs).
    ami = {
        'amazon-linux': 'ami-146e2a7c',
        'redhat':       'ami-12663b7a',
        'suse':         'ami-aeb532c6',
        'ubuntu':       'ami-9a562df2',
    }

    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Determine whether to use a start-up AMI or a specific AMI.
    if image_id:
        image = ec2_connection.get_image(image_id)
        if not image:
            raise RuntimeError('The specified Amazon Machine Image (AMI) could not be found (%s).' % image_id)
    else:
        image_id = ami[os]

    # Generate EC2 Instance name, if one was not specified.
    if not name:
        random_id = '{:08x}'.format(random.randrange(2**32))
        name = '-'.join(['ec2', config['PROJECT_NAME'], config['ENVIRONMENT'], random_id])

    # Generate Elastic Network Interface (ENI) name.
    eni_name = '-'.join(['eni', name.replace('ec2-', '')])

    # Create Elastic Network Interface (ENI) specification.
    interface = boto.ec2.networkinterface.NetworkInterfaceSpecification(subnet_id=subnet.id,
                                                                        groups=[security_group.id for security_group in security_groups],
                                                                        associate_public_ip_address=internet_addressable)
    interfaces = boto.ec2.networkinterface.NetworkInterfaceCollection(interface)

    # Create EC2 Reservation.
    logger.info('Creating EC2 Instance (%s) in %s.' % (name, subnet.availability_zone))
    reservation = ec2_connection.run_instances(image_id,                 # image_id
                                               instance_type='t2.micro',
                                               instance_profile_name=instance_profile.name if instance_profile else None,
                                               network_interfaces=interfaces,
                                               user_data=script)
    logger.info('Created EC2 Instance (%s).' % name)

    # Get EC2 Instances.
    instances = [instances for instances in reservation.instances]

    # Tag EC2 Instances.
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([instance.id for instance in instances], {'Name': name,
                                                                                          'Project': config['PROJECT_NAME'],
                                                                                          'Environment': config['ENVIRONMENT'],
                                                                                          'Role': role if role else '',})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidInstanceID.NotFound': # Instance hasn't registered with EC2 service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    # Get Elastic Network Interface (ENI) attached to instances.
    interfaces = None
    while not interfaces:
        try:
            interfaces = ec2_connection.get_all_network_interfaces(filters={'attachment.instance-id': [instance.id for instance in instances]})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidInstanceID.NotFound': # Instance hasn't registered with EC2 service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    # Tag Elastic Network Interface (ENI).
    tagged = False
    while not tagged:
        try:
            tagged = ec2_connection.create_tags([interface.id for interface in interfaces], {'Name': eni_name,
                                                                                             'Project': config['PROJECT_NAME'],
                                                                                             'Environment': config['ENVIRONMENT']})
        except boto.exception.EC2ResponseError as error:
            if error.code == 'InvalidNetworkInterfaceID.NotFound': # ENI hasn't registered with EC2 service yet.
                pass
            else:
                raise boto.exception.EC2ResponseError

    return instances

def get_nat_image(paravirtual=False):
    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Get paravirtual (PV) or hardware virtual machine (HVM) Amazon Linux VPC NAT AMIs.
    images = ec2_connection.get_all_images(filters={'owner-alias': 'amazon',
                                                    'name': 'amzn-ami-vpc-nat-' + ('pv' if paravirtual else 'hvm') + '*',})

    # Return the most recent AMI.
    image = sorted(images, key=lambda x: x.name.split('-')[5])[-1]
    return image

def run(script, command):
    script += '\n' + command
    return script

def install_package(script, package_name):
    script += '\n' + 'apt-get --yes --quiet install %s' % package_name
    return script

def register_instances(load_balancer, servers):
    # Connect to the Amazon EC2 Load Balancing (Amazon ELB) service.
    logger.debug('Connecting to the Amazon EC2 Load Balancing (Amazon ELB) service.')
    elb_connection = boto.connect_elb()
    logger.debug('Connected to the Amazon EC2 Load Balancing (Amazon ELB) service.')

    elb_connection.register_instances(load_balancer.name, [server.id for server in servers])

def get_instances(role=None):
    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()

    # Set up filters.
    filters = {}
    filters['tag:Project'] = config['PROJECT_NAME']
    filters['tag:Environment'] = config['ENVIRONMENT']
    if role:
        filters['tag:Role'] = role

    # Get reservations by tag.
    reservations = ec2_connection.get_all_instances(filters=filters)
    
    # Get instances from reservations.
    instances = [instance for reservation in reservations for instance in reservation.instances]
    
    return instances
    
def terminate_instances(instances):
    # Connect to the Amazon Elastic Compute Cloud (Amazon EC2) service.
    ec2_connection = connect_ec2()
    
    # Terminate EC2 instances.
    if instances:
        ec2_connection.terminate_instances(instance_ids=[instance.id for instance in instances])
